#include <iostream>
#include <fstream>
#include <algorithm>
#include "temporalInstance.h"
#include "lexScanner.h"
using namespace std;


//*****************************************************************************************
//*****************************************************************************************
//The maximum number of variables allowed in temporal reasoning
//const int MAX_NUM_OF_VARIABLES = 16;
//
//*****************************************************************************************
// The State class
//*****************************************************************************************
//class State
//{
//	private:
//		// Internally, a state object uses a bitset to record the binary values of the state variables.
//		// The balue of bit i records the value of state varibale X_i in this particular state   
//		bitset<MAX_NUM_OF_VARIABLES> VariableValues;
//
//	public:
//		//Set the values of the  variables according toth egiven bitset
//		void SetState(const bitset<MAX_NUM_OF_VARIABLES> & aBitset);
//
//		//Set the values of the selected variables to 1; Return true if done successfully
//		bool SetState(const vector<unsigned short> & vectOfVarIds);
//
//		//Set values of all variables to 1
//		void SetAll();
//
//		//Reset the values of the selected variables to 0; Return true if done successfully
//		bool ResetState(const vector<int> & vectOfVarIds);
//
//		//Reset values of all variables to 0
//		void ResetAll();
//
//		//display the state
//		void Display(const map<unsigned int, string> & variableIdToNameMap) const;
//
//		//write the state information to a file
//		void WriteToFile(const map<unsigned int, string> & variableIdToNameMap, ofstream & fout) const;
//
//		//comparing two states
//		inline bool operator <(const State  & anotherState) const
//		{ return ( (this->VariableValues.to_ulong() ) < anotherState.VariableValues.to_ulong() ); };
//
//		//comparing two states
//		inline bool operator ==(const State  & anotherState) const
//		{ return ( (this->VariableValues) == anotherState.VariableValues ); };
//
//		//comparing two states
//		inline bool operator !=(const State  & anotherState) const
//		{ return ( (this->VariableValues) != anotherState.VariableValues ); };
//
//		//Get the bitset representation of the state
//		bitset<MAX_NUM_OF_VARIABLES>  GetVariableBitset() const;
//
//	//The StateExpression class can access private members of State
//	friend class StateExpression;
//
//	//The CausalRule class can access private members of State
//	friend class CausalRule;
//
//	//TemporalInstance class can access private members of State
//	friend class TemporalInstance;
//
//};


void State::SetState(const bitset<MAX_NUM_OF_VARIABLES> & aBitset)
{
	VariableValues = aBitset;
	return;
}

bool State::SetState(const vector<unsigned short> & vectOfVarIds)
{
	for (unsigned short i=0; i < vectOfVarIds.size(); i++)
		if (	vectOfVarIds[i] < 0 || 
				vectOfVarIds[i] >= MAX_NUM_OF_VARIABLES
		   )
			{	cout<< "Wrong variable id! At most " << MAX_NUM_OF_VARIABLES 
					<< " variables can be used" <<endl;
				return false;
			}
		else
            VariableValues.set(vectOfVarIds[i]);

	return true;
}

void State::SetAll()
{
	VariableValues.set();
}

bool State::ResetState(const vector<int> & vectOfVarIds)
{
	for (unsigned int i=0; i < vectOfVarIds.size(); i++)
		if (	vectOfVarIds[i] < 0 ||
				vectOfVarIds[i] >= MAX_NUM_OF_VARIABLES
		   )
			{	cout<< "Wrong variable id! At most " << MAX_NUM_OF_VARIABLES 
					<< " variables can be used" <<endl;
				return false;
			}
		else
			VariableValues.reset(vectOfVarIds[i]);

	return true;
}

void State::ResetAll()
{
	VariableValues.reset();
}

void State::Display(const map<unsigned int, string>& variableIdToNameMap) const
{
	for (unsigned int i=0; i<VariableValues.size(); i++)
		if (variableIdToNameMap.find(i) != variableIdToNameMap.end() )
			//cout << '\t' << variableIdToNameMap[i] << "\t= " << VariableValues[i] << endl;
			cout << '\t' << variableIdToNameMap.find(i)->second << "\t= " << VariableValues[i] << endl;
		else
			cout << '\t' << "Unused variable #" << i << "\t= " << VariableValues[i] << endl;
}

void State::WriteToFile
(const map<unsigned int, string> & variableIdToNameMap, ofstream & fout) const
{
	for (unsigned int i=0; i<VariableValues.size(); i++)
		if (variableIdToNameMap.find(i) != variableIdToNameMap.end() )
			//cout << '\t' << variableIdToNameMap[i] << "\t= " << VariableValues[i] << endl;
			fout << '\t' << variableIdToNameMap.find(i)->second << "\t= " << VariableValues[i] << endl;
		else
			fout << '\t' << "Unused variable #" << i << "\t= " << VariableValues[i] << endl;
}


bitset<MAX_NUM_OF_VARIABLES> State::GetVariableBitset() const
{
	return VariableValues;
}


//*****************************************************************************************
// The StateExpression class
//*****************************************************************************************		
//class StateExpression
//{ 
//	private:
//		//Information of the positive variables, which must be true in this state expression.
//		//	The corresponding bits of theses variables are set to 1 in this bitset.
//		bitset<MAX_NUM_OF_VARIABLES> PositiveConditions;
//
//
//		//Information of the negative variables, which must be false in this state expression.
//		//	The corresponding bits of theses variables are set to 1 in this bitset
//		bitset<MAX_NUM_OF_VARIABLES> NegativeConditions;
//
//	public:
//		//Determine wehther the expression if true in a give state; return true if so
//		bool IsTrueIn(const State & thisState) const;
//
//		//Set the two variable bitsets according to the information of the IDs of
//		//	the positive or negative variables respectively.
//		bool SetStateExpression
//		(	const vector<unsigned short> & vectOfVarIdsOfPositiveConditions, 
//			const vector<unsigned short> & vectOfVarIdsOfNegativeConditions	);
//
//		//display a state expression with respect to a map from variable IDs to names
//		void Display(const map<unsigned int, string> & VariableIdToNameMap) const;
//
//		//write to file a state expression with respect to a map from variable IDs to names
//		void WriteToFile(const map<unsigned int, string> & VariableIdToNameMap, ofstream & fout) const;
//
//	//The CausalRule class can access private members of StateExpression
//	friend class CausalRule;
//
//	//TemporalInstance class can access private members of StateExpression
//	friend class TemporalInstance;
//
//
//};



bool StateExpression::IsTrueIn(const State & thisState) const
{
	bitset<MAX_NUM_OF_VARIABLES> test;

    test = thisState.GetVariableBitset();
	test &= PositiveConditions;
	test ^= PositiveConditions;
	if ( test.any() )
		return false;

    test = thisState.GetVariableBitset();
	test.flip();
	test &= NegativeConditions;
	test ^= NegativeConditions;
	if ( test.any() )
		return false;

	return true;
}



bool StateExpression::SetStateExpression
(	const vector<unsigned short> & vectOfVarIdsOfPositiveConditions, 
	const vector<unsigned short> & vectOfVarIdsOfNegativeConditions	
)
{
	PositiveConditions.reset();
	for (unsigned int i=0; i < vectOfVarIdsOfPositiveConditions.size(); i++)
	{	if (	vectOfVarIdsOfPositiveConditions[i] <0 || 
				vectOfVarIdsOfPositiveConditions[i] >= MAX_NUM_OF_VARIABLES
		   )
		{	cout<< "Wrong variable id! At most " << MAX_NUM_OF_VARIABLES 
				<< " variables can be used" <<endl;
			return false;
		}
		PositiveConditions.set(vectOfVarIdsOfPositiveConditions[i]);
	}

	NegativeConditions.reset();
	for (unsigned int i=0; i < vectOfVarIdsOfNegativeConditions.size(); i++)
	{	if (	vectOfVarIdsOfNegativeConditions[i] <0 ||
				vectOfVarIdsOfNegativeConditions[i] >= MAX_NUM_OF_VARIABLES
		   )
		{	cout<< "Wrong variable id! At most " << MAX_NUM_OF_VARIABLES 
				<< " variables can be used" <<endl;
			return false;
		}
		NegativeConditions.set(vectOfVarIdsOfNegativeConditions[i]);
	}

	return true;

}


//display a state expression with respect to a map from variable IDs to names
void StateExpression::Display(const map<unsigned int, string> & VariableIdToNameMap) const
{
	int j=0;
	cout << "\t\t" ;
	for (	unsigned int i = 0; i < PositiveConditions.size(); i++ )
	{	
		if ( PositiveConditions[i]  )
		{
			cout  << VariableIdToNameMap.find(i)->second << ',';
			if ( (++j % 3 ) == 0 )
				cout << endl << "\t\t";
		}
	}

	for (	unsigned int i = 0; i < NegativeConditions.size(); i++ )
	{	
		if ( NegativeConditions[i]  )
		{
			cout  << "not " << VariableIdToNameMap.find(i)->second << ',';
			if ( (++j % 3 ) == 0 )
				cout << endl << "\t\t";
		}
	}
}


//write to file a state expression with respect to a map from variable IDs to names
void StateExpression::WriteToFile
(const map<unsigned int, string> & VariableIdToNameMap, ofstream & fout) const
{
	int j=0;
	fout << "\t\t" ;
	for (	unsigned int i = 0; i < PositiveConditions.size(); i++ )
	{	
		if ( PositiveConditions[i]  )
		{
			fout  << VariableIdToNameMap.find(i)->second << ',';
			if ( (++j % 3 ) == 0 )
				fout << endl << "\t\t";
		}
	}

	for (	unsigned int i = 0; i < NegativeConditions.size(); i++ )
	{	
		if ( NegativeConditions[i]  )
		{
			fout  << "not " << VariableIdToNameMap.find(i)->second << ',';
			if ( (++j % 3 ) == 0 )
				fout << endl << "\t\t";
		}
	}
}

//*****************************************************************************************
//The CausalRule class
//*****************************************************************************************
//class CausalRule
//{ 
//	private:
//		//The preconditions represented as a StateExpression
//		StateExpression PreConditions;
//
//		//The postconditions represented as a StateExpression
//		StateExpression PostConditions;
//
//	public:
//		//Get the next state from thisState 
//		State GetNextStateFrom(const State & thisState) const;
//
//		//Set up the causal rule representation; Return true if successfully done
//		bool SetCausalRule
//		(	const vector<unsigned short> & vectOfVarIdsOfPositivePreConditions, 
//			const vector<unsigned short> & vectOfVarIdsOfNegativePreConditions,	
//			const vector<unsigned short> & vectOfVarIdsOfPositivePostConditions, 
//			const vector<unsigned short> & vectOfVarIdsOfNegativePostConditions	
//		);
//
//		//display a causal rule with respect to a map from variable IDs to names
//		void Display(const map<unsigned int, string> & VariableIdToNameMap) const;
//
//		//write to a file a causal rule with respect to a map from variable IDs to names
//		void WriteToFile
//			(const map<unsigned int, string> & VariableIdToNameMap, ofstream & fout) const;
//
//		//TemporalInstance class can access private members of CausalRules
//	friend class TemporalInstance;
//};


State CausalRule::GetNextStateFrom(const State & thisState) const
{
	if ( PreConditions.IsTrueIn(thisState) == false )
		return thisState;

	bitset<MAX_NUM_OF_VARIABLES> nextStateBitset;
	nextStateBitset = thisState.GetVariableBitset();
	nextStateBitset |= PostConditions.PositiveConditions;

	bitset<MAX_NUM_OF_VARIABLES> mask;
	mask = PostConditions.NegativeConditions;
	mask.flip();
	nextStateBitset &= mask;

	State nextState;
	nextState.SetState(nextStateBitset);
	return nextState;
}

bool CausalRule::SetCausalRule
(	const vector<unsigned short> & vectOfVarIdsOfPositivePreConditions, 
	const vector<unsigned short> & vectOfVarIdsOfNegativePreConditions,	
	const vector<unsigned short> & vectOfVarIdsOfPositivePostConditions, 
	const vector<unsigned short> & vectOfVarIdsOfNegativePostConditions	
)
{
	return 
	PreConditions.SetStateExpression
		(vectOfVarIdsOfPositivePreConditions,vectOfVarIdsOfNegativePreConditions) 
	&&
	PostConditions.SetStateExpression
		(vectOfVarIdsOfPositivePostConditions,vectOfVarIdsOfNegativePostConditions);

}

void CausalRule::Display(const map<unsigned int, string> & VariableIdToNameMap) const
{		
		cout << '\t' << "If before the event we have" << endl;
		PreConditions.Display(VariableIdToNameMap);
		cout << endl << '\t' << "then after the event we might have" << endl;
		PostConditions.Display(VariableIdToNameMap);
		cout << endl;
};


void CausalRule::WriteToFile
(const map<unsigned int, string> & VariableIdToNameMap, ofstream & fout) const
{
		fout << '\t' << "If before the event we have" << endl;
		PreConditions.WriteToFile(VariableIdToNameMap, fout);
		fout << endl << '\t' << "then after the event we might have" << endl;
		PostConditions.WriteToFile(VariableIdToNameMap, fout);
		fout << endl;
}

//*****************************************************************************************
// The Event class
//*****************************************************************************************


//class Event
//{
//	public:
//		//Define double as the type for point-of-time information
//		typedef double TimePoint;
//
//	private:
//		//Name of the event
//		string EventName;
//
//		//The causal rules
//		vector<CausalRule> CausalRules;
//
//		//Must occur after this point of time
//		TimePoint OccurAfterThisTime;
//
//		//Must occur before this point of time
//		TimePoint OccurBeforeThisTime;
//
//	public:
//
//		//display an event with respect to a map from variable IDs to names
//		void Display(map<unsigned int, string> & VariableIdToNameMap);
//
//		//Add a new causal rule
//		void AddCausalRule(CausalRule & aRule);
//
//		//Set the name
//		void SetName(string aName);
//
//		//Set the two time points OccurAfterThisTime and OccurBeforeThisTime
//		void SetTime(TimePoint OA, TimePoint OB);
//
//		//Determine whether one event can only strictly occur after another
//		bool operator < (Event & anoterEvent);
//
//		//Determine whether one event can only strictly occur before another
//		bool operator > (Event & anoterEvent);
//
//		//Determine whether the time intervals of two events overlap
//		bool Overlap(Event & anoterEvent);
//
//	//TemporalInstance class can access private members of events
//	friend class TemporalInstance;
//
//	//TemporalSolution class can access private members of events
//	friend class TemporalProjection;
//
//};


//display an event with respect to a map from variable IDs to names
void Event::Display(map<unsigned int, string> & VariableIdToNameMap)
{
	cout	<< endl << endl
			<< "***************  Event " << EventName << " ****************" << endl
			<< "Event " << EventName 
			<< " shall occur at a point of time in the time interval [" 
			<< OccurAfterThisTime << ", "  
			<< OccurBeforeThisTime << "]." <<endl;
	cout	<< "Event " << EventName 
			<< " has " << CausalRules.size() << " causal rule(s):" << endl;

	for (unsigned int i=0; i<CausalRules.size(); i++)
	{	cout	<< endl
				<< '\t' << "******* Rule " << i << " *******" << endl;
		CausalRules[i].Display(VariableIdToNameMap);
	}

};

void Event::SetName(string aName)
{
	EventName = aName;
}

void Event::SetTime(TimePoint OA, TimePoint OB)
{
	OccurAfterThisTime = OA;
	OccurBeforeThisTime = OB;
}

void Event::AddCausalRule(CausalRule & aRule)
{
	CausalRules.push_back(aRule);
}

bool Event::operator < (Event & anotherEvent)
{
	return this->OccurBeforeThisTime <= anotherEvent.OccurAfterThisTime;
}

bool Event::operator > (Event & anotherEvent)
{
	return this->OccurAfterThisTime >= anotherEvent.OccurBeforeThisTime;
}

bool Event::Overlap(Event & anotherEvent)
{
	return 
		( !( (*this) > anotherEvent )  && !( (*this) < anotherEvent ) );
}


//*****************************************************************************************
//The TemporalInstance class 
//*****************************************************************************************
//class TemporalInstance
//{
//	private:
//		//Instance name
//		string InstanceName;
//
//		//Initial state
//		State InitialState;
//
//		//Goal
//		StateExpression Goal;
//
//		//A map for mapping the names of the events into their IDs
//		map<string, unsigned short> VariableNameToIdMap;
//
//		//A map for mapping the IDs of the events into their names
//		map<unsigned int, string> VariableIdToNameMap;
//
//		//The events
//		vector<Event> Events;
//
//		//Read the lines of description from a file into a vector of strings
//		//	a privatemethod called by 
//		bool ReadLinesFromFile(char * fileName, vector<string> &linesOfDescription);
//
//	public:
//
//		//Check whether the time interval information is consistent
//		bool IsTimeInformationConsistent() const;
//
//		//Read the contents of a temporal instance from a file
//		bool ReadInstanceFromFile(char * fileName);
//
//		//Display the contents of a temporal instance
//		void Display();
//
//
//	//The TemporalProjection class is allowed to acccess the prvate members of TemporalInstance
//	friend class TemporalProjection;
//
//};

bool  TemporalInstance::IsTimeInformationConsistent() const
{
	bool isConsistent = true;
	vector<Event::TimePoint> timePointVect;
	for (unsigned int i =0; i<Events.size(); i++)
	{
		if ( Events[i].OccurAfterThisTime >= Events[i].OccurBeforeThisTime )
		{
			cout << "Event " << Events[i].EventName 
				 <<": Inconsistent time interval information" << endl;
			isConsistent = false;
		}
	}
	return isConsistent;
}

bool TemporalInstance::ReadLinesFromFile
(char * fileName, vector<string> &linesOfDescription)
{
    ifstream inFile;
    inFile.open( fileName );
	if ( inFile.fail() )
	{ cout << endl << "Cannot open this file!" << endl;
	  return false;
	}
	
	linesOfDescription.clear();  //Delete all the lines of statements
	string inputBuffer;
    while( !inFile.eof() )
    { 
        getline(inFile,inputBuffer) ;
        while( !inFile.eof() && inputBuffer.size() == 0 )
        {
            getline(inFile, inputBuffer);
        }
		if (!inFile.eof())
		{
  			cout << "Line[" 
				<< linesOfDescription.size() 
				<<"] read:" << '\t' 
				<< inputBuffer << endl;
			linesOfDescription.push_back(inputBuffer);
		}

    }
	
    inFile.close();
	return true;

}


bool TemporalInstance::ReadInstanceFromFile(char * fileName)
{
	vector<string> linesOfDescription;
	if ( ReadLinesFromFile(fileName, linesOfDescription) == false )
		return false;

	LexScanner myLexScanner;
	LexScanner::VectOfTokenVects		tokenVectors;
	LexScanner::VectOfCategoryVects		categoryVectors;

	myLexScanner.getLexicalInfoAllLines(linesOfDescription, tokenVectors, categoryVectors);
	//myLexScanner.displayLexicalInfo(tokenVectors, categoryVectors);

	Event thisEvent;
	CausalRule thisRule;
	vector<unsigned short> vectOfVarIdsOfPositivePreConditions; 
	vector<unsigned short> vectOfVarIdsOfNegativePreConditions;	
	vector<unsigned short> vectOfVarIdsOfPositivePostConditions;
	vector<unsigned short> vectOfVarIdsOfNegativePostConditions;	

	for (unsigned int i=0; i<tokenVectors.size(); i++)
	{

		LexScanner::PerLineTokenVector tokenVect;
		LexScanner::PerLineTokenCategoryVector categoryVect;

		tokenVect = tokenVectors[i];
		categoryVect = categoryVectors[i];
		
		if (tokenVect.size() != categoryVect.size() )
		{	cout << "Error in the lexical module." << endl;
			return false;
		}



		//************************************************
		//CASE: Declaration of variables
		//************************************************
		if (	tokenVect.size() >= 2			&&
				(tokenVect[0] == "Variables"	 || tokenVect[0] == "variables"	)
			)
		{
			//Set up the loop to record the names of new variables
			for (unsigned int i=1; i<tokenVect.size(); i++)
			{
				if ( categoryVect[i] == LexScanner::ID_NAME &&
					 VariableNameToIdMap.find(tokenVect[i]) == VariableNameToIdMap.end() 
				   )
				{	unsigned int ID = VariableNameToIdMap.size();
					VariableNameToIdMap[ tokenVect[i] ] = ID;
				}
			}
		}

		//*****************************************************
		//CASE: Initial State with the "Initial State" keywords
		//*****************************************************
		if (	tokenVect.size() >= 3				&&
				(tokenVect[0] == "Initial"	|| tokenVect[0] == "initial")	&&	
				(tokenVect[1] == "State"	|| tokenVect[1] == "state" )
			)
		{

			vector<unsigned short> vectOfPositiveVarIds; 
			vector<unsigned short> vectOfNegativeVarIds;	
			
			//Iteratively add precondition information
			bool isPositiveVar;
			isPositiveVar = true;
			for (unsigned int i=2; i < tokenVect.size(); i++)
			{
				if (tokenVect[i] == "not" ||tokenVect[i] == "Not")
					isPositiveVar = false;
				if (categoryVect[i] == LexScanner::ID_NAME )
				{	unsigned int varId;
					if (	VariableNameToIdMap.find(tokenVect[i]) == 
							VariableNameToIdMap.end()
					   )
					{	unsigned int ID;
						ID = VariableNameToIdMap.size();
						VariableNameToIdMap[tokenVect[i]] = ID;
					}

					varId = VariableNameToIdMap[tokenVect[i]] ;

					if (isPositiveVar)
						vectOfPositiveVarIds.push_back(varId);
					else
						vectOfNegativeVarIds.push_back(varId);

					isPositiveVar = true;
				}

			}

			//Reset the values of all state variables of the initial state to 0
			InitialState.ResetAll();
			//Set the values of all state variables of the initial state to 1
			InitialState.SetState(vectOfPositiveVarIds);
			
		}

		//*****************************************************
		//CASE: Goal eclaration with the "Goal" keyword
		//*****************************************************
		if (	tokenVect.size() >= 2				&&
				(tokenVect[0] == "Goal"	|| tokenVect[0] == "goal")	
			)
		{

			vector<unsigned short> vectOfPositiveVarIds; 
			vector<unsigned short> vectOfNegativeVarIds;	
			
			//Iteratively add precondition information
			bool isPositiveVar;
			isPositiveVar = true;
			for (unsigned int i=2; i < tokenVect.size(); i++)
			{
				if (tokenVect[i] == "not" ||tokenVect[i] == "Not")
					isPositiveVar = false;
				if (categoryVect[i] == LexScanner::ID_NAME )
				{	int varId;
					if (	VariableNameToIdMap.find(tokenVect[i]) == 
							VariableNameToIdMap.end()
					   )
					{	int ID;
						ID = VariableNameToIdMap.size();
						VariableNameToIdMap[tokenVect[i]] = ID;
					}

					varId = VariableNameToIdMap[tokenVect[i]] ;

					if (isPositiveVar)
						vectOfPositiveVarIds.push_back(varId);
					else
						vectOfNegativeVarIds.push_back(varId);

					isPositiveVar = true;
				}

			}

			//Set the goal state
			Goal.SetStateExpression(vectOfPositiveVarIds, vectOfNegativeVarIds);
			
		}


		//************************************************
		//CASE: Declaration of events
		//************************************************
		if (	tokenVect.size() >= 2			&&
				(tokenVect[0] == "Event"	 || tokenVect[0] == "event"	)
			)
		{
			//Store the information of the current event
			if ( ! thisEvent.CausalRules.empty() )
				Events.push_back(thisEvent);

			//Start a new event and cear up the causal rules of last event
			thisEvent.CausalRules.clear(); 

			//Record the name of the event
			thisEvent.EventName = tokenVect[1];

			//Figure out OccurAfterThisTime and OccurBeforeThisTime
			int i = 1;
			for ( ; i < tokenVect.size() &&  categoryVect[i]!= LexScanner::NUMERICAL_LITERAL; i++) 
				{};
			if (i == tokenVect.size() )
				cout << "Event " << tokenVect[1] << " missing the first time interval information!!" << endl;
			else
				thisEvent.OccurAfterThisTime = atof( tokenVect[i].c_str());

			for (++i ; i < tokenVect.size() &&  categoryVect[i]!= LexScanner::NUMERICAL_LITERAL; i++) 
				{};
			if (i == tokenVect.size() )
				cout << "Event " << tokenVect[1] << " missing the second time interval information!!" << endl;
			else
				thisEvent.OccurBeforeThisTime = atof( tokenVect[i].c_str() );
		}


		//************************************************
		//CASE: Preconditions of a causal rule
		//************************************************
		if (	tokenVect.size() >= 2									&&
				( tokenVect[0] == "If"	|| 	tokenVect[0] == "if")		
			)
		{

			//Before starting a new rule,
			//Clear the preconditions and postconditions to start a new rule
			vectOfVarIdsOfPositivePreConditions.clear();
			vectOfVarIdsOfNegativePreConditions.clear();	
			vectOfVarIdsOfPositivePostConditions.clear();
			vectOfVarIdsOfNegativePostConditions.clear();

			//Iteratively add precondition information
			bool isPositiveVar;
			isPositiveVar = true;
			for (int i=1; i < tokenVect.size(); i++)
			{
				if (tokenVect[i] == "not" ||tokenVect[i] == "Not")
					isPositiveVar = false;
				if (categoryVect[i] == LexScanner::ID_NAME )
				{	//Record the variable name if this is the first time it appears
					if (	VariableNameToIdMap.find(tokenVect[i]) == 
							VariableNameToIdMap.end()
					   )
					{	int ID;
						ID = VariableNameToIdMap.size();
						VariableNameToIdMap[tokenVect[i]] = ID;
					}

					int varId;
					varId = VariableNameToIdMap[tokenVect[i]] ;

					if (isPositiveVar)
						vectOfVarIdsOfPositivePreConditions.push_back(varId);
					else
						vectOfVarIdsOfNegativePreConditions.push_back(varId);

					isPositiveVar = true;

				}

			}

		}

		//************************************************
		//CASE: Postconditions of a causal rule
		//************************************************
		if (	tokenVect.size() >= 2									&&
				( tokenVect[0] == "then"	|| 	tokenVect[0] == "Then")		
			)
		{
			//Iteratively add postcondition information
			bool isPositiveVar;
			isPositiveVar = true;
			for (int i=1; i < tokenVect.size(); i++)
			{
				if (tokenVect[i] == "not" ||tokenVect[i] == "Not")
					isPositiveVar = false;
				if (categoryVect[i] == LexScanner::ID_NAME )
				{
					if (	VariableNameToIdMap.find(tokenVect[i]) == 
							VariableNameToIdMap.end()
					   )
					{	int ID;
						ID = VariableNameToIdMap.size();
						VariableNameToIdMap[tokenVect[i]] = ID;
					}
					int varId;
					varId = VariableNameToIdMap[tokenVect[i]] ;

					if (isPositiveVar)
						vectOfVarIdsOfPositivePostConditions.push_back(varId);
					else
						vectOfVarIdsOfNegativePostConditions.push_back(varId);

					isPositiveVar = true;

				}
			}

			//Add thisRule to thisEvent if thisRule is not an empty rule
			if ( 	!vectOfVarIdsOfPositivePreConditions.empty()	|| 
					!vectOfVarIdsOfNegativePreConditions.empty()	||	
					!vectOfVarIdsOfPositivePostConditions.empty()	||
					!vectOfVarIdsOfNegativePostConditions.empty()
			   )
			{
				thisRule.SetCausalRule
					(	vectOfVarIdsOfPositivePreConditions, 
						vectOfVarIdsOfNegativePreConditions,	
						vectOfVarIdsOfPositivePostConditions, 
						vectOfVarIdsOfNegativePostConditions	
					);
				thisEvent.CausalRules.push_back(thisRule);
			}

			//Clear the condition information to start a new rule
			vectOfVarIdsOfPositivePreConditions.clear();
			vectOfVarIdsOfNegativePreConditions.clear();	
			vectOfVarIdsOfPositivePostConditions.clear();
			vectOfVarIdsOfNegativePostConditions.clear();
		}


	}

	//************************************************
	//CASE: After all information is read and processed
	//************************************************
	//Add the very last event into the collection of events
	if ( ! thisEvent.CausalRules.empty() )
		Events.push_back(thisEvent);

	//************************************************
	//Sort the events according to the starting time
	//************************************************
	sort(this->Events.begin(), this->Events.end(), 
			Compare_Events_By_Their_EarliestStartingPoints() );

	//************************************************
	//construct a map from variable IDs to variable names
	//************************************************
	for (	map<string, unsigned short>::iterator itr = VariableNameToIdMap.begin();
			itr != VariableNameToIdMap.end(); 
			itr++
		)
		VariableIdToNameMap[itr->second] = itr->first;


	return true;
}


//Display the contents of a temporal instance
void TemporalInstance::Display() 
{	//Display names of all variables
	int i = 0;
	cout << endl << endl
		 << "*************  Variables  ***************" << endl
		 << "List of all " << VariableNameToIdMap.size() << " variables in order: " << endl;

	for (	map<string, unsigned short>::iterator itr = VariableNameToIdMap.begin();
			itr != VariableNameToIdMap.end(); 
			itr++
		)
	{
		cout << '\t' << "Variable#" << itr->second << ": " <<  itr->first;
		if ( ( (++i) % 3 ) == 0 )
			cout << endl;
	}
	if (MAX_NUM_OF_VARIABLES > VariableNameToIdMap.size() )
	{
		cout << endl << endl
			 << "This program can handle up to MAX_NUM_OF_VARIABLES=" 
			 << MAX_NUM_OF_VARIABLES
			 << " variables." 
			 << endl
			 << "So there are " 
			 << MAX_NUM_OF_VARIABLES - VariableNameToIdMap.size()
			 << " extra unused variables."
			 << endl;
	}


	//Display the initial state
	cout << endl << endl
		 << "*************  Initial state  ***************" << endl; 
	InitialState.Display(VariableIdToNameMap); cout << endl << endl;

	//Display the goal
	cout	<< endl << endl
			<< "*****************************************************" << endl
			<< "Goal: " << endl;
	Goal.Display(VariableIdToNameMap);


	//Display the events and their associated rules
	cout	<< endl << endl;
	cout	<< "*****************************************************" << endl
			<< "List of all "<< Events.size() << " events and their associated causal rules " 
			<< endl;

	for (	unsigned int i=0; i < Events.size(); i++ )
		Events[i].Display(VariableIdToNameMap);

	cout	<< endl 
			<< "*****************************************************" << endl
			<< endl << endl;

}

